"""
 DEX SWAP
🛠️ W3 UTILS
"""

import requests
import aiohttp
from loguru import logger

MAX_RESPONSE_SIZE = 5 * 1024 * 1024  # Maximum response size in bytes (e.g., 5 MB)

async def get(url, params=None, headers=None):
    """
    Asynchronously gets a url payload
    and returns the response

    Args:
        url (str): The url to get
        params (dict, optional): The params to send. Defaults to None.
        headers (dict, optional): The headers to send. Defaults to None.

    Returns:
        dict: The response

    Raises:
        Exception: Error

    """
    
    # try:
    #     response = requests.get(url, params=params, headers=headers, timeout=20)
    #     logger.debug(response)
    #     if response.status_code == 200:
    #         return response.json()

    # except Exception as error:
    #     logger.error("get: {}", error)

    try:
        async with aiohttp.ClientSession() as session:
            async with session.get(
             url,
             params=None,
             headers=None, 
             timeout=20) as response:
                logger.debug(response)
                if response.status == 200:
                    if response.content_length > MAX_RESPONSE_SIZE:
                        logger.warning("Response content too large, skipping...")
                        return None  # Return None for large responses
                    return await response.json(content_type=None)

    except Exception as error:
        logger.error("get: {}", error)




"""
 DEX SWAP
EXPLORER
"""

from datetime import datetime, timedelta

from loguru import logger

from dxsp.config import settings
from dxsp.utils.utils import get


async def get_explorer_abi(address):
    """
    Retrieves the ABI (Application Binary Interface)
    for the contract at the given address.

    :param address: The address of the contract.
    :type address: str

    :return: The ABI of the contract if it exists, else None.
    :rtype: str or None
    """
    if not settings.dex_block_explorer_api:
        return None

    params = {
        "module": "contract",
        "action": "getabi",
        "address": address,
        "apikey": settings.dex_block_explorer_api,
    }
    resp = await get(url=settings.dex_block_explorer_url, params=params)
    return resp["result"] if resp["status"] == "1" else None


async def get_account_transactions(contract_address, wallet_address, period=24):
    """
    Retrieves the account transactions
    within a specified time period
    for the main asset activity
    Not yet implemented

    :param contract_address: The address of the contract.
    :type contract_address: str
    :param wallet_address: The address of the wallet.
    :type wallet_address: str
    :param period: The time period in hours
    :type period: int

    :return: The transactions for the account.
    """
    pnl_dict = {"pnl": 0, "tokenList": {}}
    if not settings.dex_block_explorer_api:
        return pnl_dict

    params = {
        "module": "account",
        "action": "tokentx",
        "contractaddress": contract_address,
        "address": wallet_address,
        "page": "1",
        "offset": "100",
        "startblock": "0",
        "endblock": "99999999",
        "sort": "desc",
        "apikey": settings.dex_block_explorer_api,
    }

    response = await get(url=settings.dex_block_explorer_url, params=params)

    if response.get("status") == "1" and "result" in response:
        current_time = datetime.utcnow()
        time_history_start = current_time - timedelta(hours=period)

        for entry in response["result"]:
            token_symbol = entry.get("tokenSymbol")
            value = int(entry.get("value", 0))
            timestamp = int(entry.get("timeStamp", 0))
            transaction_time = datetime.utcfromtimestamp(timestamp)

            if transaction_time >= time_history_start and token_symbol:
                pnl_dict["tokenList"][token_symbol] = (
                    pnl_dict["tokenList"].get(token_symbol, 0) + value
                )
                pnl_dict["pnl"] += value

    return pnl_dict



"""
 DEX SWAP Main
"""

import decimal
from typing import Optional

from loguru import logger
from web3 import Web3
from web3.gas_strategies.time_based import medium_gas_price_strategy

from dxsp.config import settings
from dxsp.utils import AccountUtils, ContractUtils


class DexSwap:
    """
    DEXswap  class to build a DexSwap Object
    use to interact with the dex protocol

    Args:
        w3 (Optional[Web3]): Web3

    Returns:
        DexSwap


    """

    def __init__(self, w3: Optional[Web3] = None):
        """
        Initialize the DexSwap object to interact with
        w3 contracts.

        """
        self.logger = logger
        self.w3 = w3 or Web3(Web3.HTTPProvider(settings.dex_rpc))
        if not self.w3.net.listening:
            raise ValueError("w3 not connected")
        self.w3.eth.set_gas_price_strategy(medium_gas_price_strategy)

        self.account = AccountUtils(w3=self.w3)

        self.protocol_type = settings.dex_protocol_type
        self.protocol_version = settings.dex_protocol_version
        self.dex_swap = None
        self.router = None
        self.quoter = None
        self.contract_utils = ContractUtils(w3=self.w3)

    async def get_protocol(self):
        """
        Set the dex_swap object based
        on the protocol type.
        Currently supports:
            1inch: currently blocked

            0x

            Uniswap V2 and V3

        Returns:
            dex_swap


        """
        from dxsp.protocols import DexSwapOneInch, DexSwapUniswap, DexSwapZeroX

        if self.protocol_type == "0x":
            self.dex_swap = DexSwapZeroX()
        elif self.protocol_type == "1inch":
            self.dex_swap = DexSwapOneInch()
        else:
            self.dex_swap = DexSwapUniswap()

    async def execute_order(self, order_params):
        """
        Execute an order function.

        Args:
            order_params (dict): The order parameters.

        Returns:
            str: The trade confirmation

        """
        try:
            self.logger.debug("execute order")
            action = order_params.get("action")
            instrument = order_params.get("instrument")
            quantity = order_params.get("quantity", 1)
            sell_token, buy_token = (
                (self.account.trading_asset_address, instrument)
                if action == "BUY"
                else (instrument, self.account.trading_asset_address)
            )
            order = await self.get_swap(sell_token, buy_token, quantity)
            if order:
                trade_confirmation = (
                    f"⬇️ {instrument}" if (action == "SELL") else f"⬆️ {instrument}\n"
                )
                trade_confirmation += order
                return trade_confirmation

        except Exception as error:
            return f"⚠️ order execution: {error}"

    async def get_swap(self, sell_token: str, buy_token: str, quantity: int) -> None:
        """
        Execute a swap

        Args:
            sell_token (str): The sell token.
            buy_token (str): The buy token.
            quantity (int): The quantity of tokens.

        Returns:
            transactionHash


        """
        try:
            self.logger.debug("get swap")
            await self.get_protocol()
            sell_token_address = sell_token
            self.logger.debug("sell token {}", sell_token_address)
            if not sell_token.startswith("0x"):
                sell_token_address = await self.contract_utils.search_contract_address(
                    sell_token
                )
            buy_token_address = buy_token
            self.logger.debug("buy token {}", buy_token_address)
            if not buy_token_address.startswith("0x"):
                buy_token_address = await self.contract_utils.search_contract_address(
                    buy_token
                )
            sell_amount = await self.contract_utils.calculate_sell_amount(
                sell_token_address, self.account.wallet_address, quantity
            )
            sell_token_amount_wei = sell_amount * (
                10 ** (await self.contract_utils.get_token_decimals(sell_token_address))
            )
            if self.protocol_type == "0x":
                await self.account.get_approve(sell_token_address)

            order_amount = int(
                sell_token_amount_wei
                * decimal.Decimal((settings.dex_trading_slippage / 100))
            )
            self.logger.debug(order_amount)
            order = await self.dex_swap.get_swap(
                sell_token_address, buy_token_address, order_amount
            )

            if not order:
                self.logger.debug("swap order error")
                raise ValueError("swap order not executed")

            signed_order = await self.account.get_sign(order)
            order_hash = str(self.w3.to_hex(signed_order))
            receipt = self.w3.wait_for_transaction_receipt(order_hash)

            if receipt["status"] != 1:
                self.logger.debug(receipt)
                raise ValueError("receipt failed")

            return await self.contract_utils.get_confirmation(
                receipt["transactionHash"]
            )

        except Exception as error:
            self.logger.debug(error)
            raise error

    async def get_quote(self, sell_token):
        """
        gets a quote for a token

        Args:
            sell_token (str): The sell token.

        Returns:
            str: The quote with the trading symbol

        """
        try:
            await self.get_protocol()
            buy_address = self.account.trading_asset_address
            sell_address = await self.contract_utils.search_contract_address(sell_token)
            quote = await self.dex_swap.get_quote(buy_address, sell_address)
            quote = f"🦄 {quote}"
            symbol = await self.contract_utils.get_token_symbol(
                self.account.trading_asset_address
            )
            return f"{quote} {symbol}"

        except Exception as error:
            return f"⚠️: {error}"

    # 🔒 USER RELATED

    async def get_info(self):
        """
        Get information from the account.

        :return: The information retrieved from the account.
        """
        return await self.account.get_info()

    async def get_help(self):
        """
        Retrieves help information
        using the `account.get_help()` method.

        :return: The help information.
        :rtype: Any
        """
        return await self.account.get_help()

    async def get_name(self):
        """
        Retrieves the name of the account.

        :return: The name of the account.
        """
        return await self.account.get_name()

    async def get_account_balance(self):
        """
        Retrieves the account balance.

        :return: The account balance.
        :rtype: float
        """
        return await self.account.get_account_balance()

    async def get_trading_asset_balance(self):
        """
        Retrieves the trading asset balance for the current account.

        :return: A dictionary containing the trading asset balance.
                 The dictionary has the following keys:
                 - 'asset': The asset symbol.
                 - 'free': The free balance of the asset.
                 - 'locked': The locked balance of the asset.
        """
        return await self.account.get_trading_asset_balance()

    async def get_account_position(self):
        """
        Retrieves the account position.

        :return: The account position.
        :rtype: AccountPosition
        """
        return await self.account.get_account_position()

    async def get_account_margin(self):
        """
        Retrieves the account margin.

        :return: The account margin.
        :rtype: float
        """
        return await self.account.get_account_margin()

    async def get_account_open_positions(self):
        """
        Retrieves the open positions of the account.

        :return: A list of open positions in the account.
        """
        return await self.account.get_account_open_positions()

    async def get_account_transactions(self, period=24):
        """
        Get the account transactions
        for a specific period.

        Args:
            period (int): The number of hours
            for which to retrieve the transactions. Defaults to 24.

        Returns:
            List[Transaction]: A list of
            transaction objects representing the account transactions.
        """
        return await self.account.get_account_transactions(period)

    async def get_account_pnl(self, period=24):
        """
        Get the profit and loss (PnL)
        for the account within a specified period.

        Args:
            period (int, optional):
            The period in hours for which to calculate the PnL.
            Defaults to 24.

        Returns:
            float: The profit and loss (PnL)
            for the account within the specified period.
        """
        return await self.account.get_account_pnl(period)



"""
 DEX SWAP CONFIG
"""
import os

from dynaconf import Dynaconf

ROOT = os.path.dirname(__file__)

settings = Dynaconf(
    envvar_prefix="TT",
    root_path=os.path.dirname(ROOT),
    settings_files=[
        os.path.join(ROOT, "default_settings.toml"),
        "talky_settings.toml",
        "settings.toml",
        ".secrets.toml",
    ],
    load_dotenv=True,
    environments=True,
    default_env="default",
)



"""
OneInch 🦄
"""
from dxsp.config import settings
from dxsp.main import DexSwap


class DexSwapOneInch(DexSwap):
    """
    A DEXSwap sub class using 1inch protocol
    https://github.com/mraniki/dxsp/issues/189

    """

    async def get_quote(self, buy_address, sell_address, amount=1):
        """
        Retrieves a quote for the
        given buy and sell addresses and amount.

        Args:
            buy_address (str): The address of the token to buy.
            sell_address (str): The address of the token to sell.
            amount (int, optional): The amount of tokens to buy. Defaults to 1.

        Returns:
            float: The quote amount rounded to 2 decimal places.
        """
        # try:
        pass
        #     min_amount = self.w3.to_wei(amount, 'ether')
        #     quote_url = (
        #         settings.dex_1inch_url
        #         + str(self.chain_id)
        #         + "/quote?fromTokenAddress="
        #         + str(buy_address)
        #         + "&toTokenAddress="
        #         + str(sell_address)
        #         + "&amount="
        #         + str(min_amount))
        #     quote_response = await self._get(
        #         url=quote_url,
        #         params=None,
        #         headers=settings.headers)
        #     self.logger.debug("quote_response {}", quote_response)
        #     if quote_response:
        #         quote_amount = quote_response['toTokenAmount']
        #         self.logger.debug("quote_amount {}", quote_amount)
        #         # quote_decimals = quote_response['fromToken']['decimals']
        #         quote = self.w3.from_wei(int(quote_amount), 'ether')
        #         # /(10 ** quote_decimals))
        #         return round(quote, 2)

    #     except Exception as error:
    #         raise ValueError(f"Approval failed {error}")

    # async def get_approve(self, token_address):
    #     # pass
    #     try:
    #         pass
    # approval_check_URL = (
    #     settings.dex_1inch_url
    #     + str(self.chain_id)
    #     + "/approve/allowance?tokenAddress="
    #     + str(sell_address)
    #     + "&walletAddress="
    #     + str(self.wallet_address))
    # approval_response = await self._get(
    #     url=approval_check_URL,
    #     params=None,
    #     headers=settings.headers)
    # approval_check = approval_response['allowance']
    # if (approval_check == 0):
    #     approval_URL = (
    #         settings.dex_1inch_url
    #         + str(self.chain_id)
    #         + "/approve/transaction?tokenAddress="
    #         + str(sell_address))
    #     approval_response = await self._get(approval_URL)
    #     return approval_response
    #     except Exception as error:
    #         raise ValueError(f"Approval failed {error}")

    # async def get_swap(self, sell_address, buy_address, amount):

    #     try:
    #         pass
    # swap_url = (
    #     settings.dex_1inch_url
    #     + str(self.chain_id)
    #     + "/swap?fromTokenAddress="
    #     + sell_address
    #     + "&toTokenAddress="
    #     + buy_address
    #     + "&amount="
    #     + amount
    #     + "&fromAddress="
    #     + self.wallet_address
    #     + "&slippage="
    #     + settings.dex_trading_slippage
    #     )
    # swap_order = await self._get(
    #     url=swap_url,
    #     params=None,
    #     headers=settings.headers
    #     )
    # swap_order_status = swap_order['statusCode']
    # if swap_order_status != 200:
    #     return
    # return swap_order
    # except Exception as error:
    #     raise ValueError(f"Swap failed {error}")



"""
0️⃣x

"""
from dxsp.config import settings
from dxsp.main import DexSwap
from dxsp.utils.utils import get


class DexSwapZeroX(DexSwap):
    """
    A DEXSwap sub class using 0x protocol
    Implementation of 0x swap protocol
    https://0x.org/docs/0x-swap-api/introduction

    """

    async def get_quote(self, buy_address, sell_address, amount=1):
        """
        Retrieves a quote for a token swap.

        Args:
            buy_address (str): The address of the token to be bought.
            sell_address (str): The address of the token to be sold.
            amount (int, optional): The amount of tokens to be sold. Defaults to 1.

        Returns:
            float: The guaranteed price for the token swap.
        """
        token_decimals = await self.contract_utils.get_token_decimals(sell_address)
        out_amount = amount * (10**token_decimals)
        url = (
            f"{settings.dex_0x_url}/swap/v1/quote"
            f"?buyToken={str(buy_address)}&sellToken={str(sell_address)}&sellAmount={str(out_amount)}"
        )
        headers = {"0x-api-key": settings.dex_0x_api_key}
        response = await get(url, params=None, headers=headers)
        if response:
            return float(response["guaranteedPrice"])

    async def get_swap(self, buy_address, sell_address, amount):
        """
        Asynchronously gets a swap order by calling the `get_quote`
        method with the specified `buy_address`,
        `sell_address`, and `amount` parameters.
        Then, it calls the `get_sign` method of the
        `account` object, passing the `swap_order`
        as an argument, and returns the result.

        :param buy_address: The buy address for the swap.
        :param sell_address: The sell address for the swap.
        :param amount: The amount for the swap.

        :return: The result of calling the `get_sign` method
        of the `account` object with the `swap_order`
        as an argument.
        """
        swap_order = await self.get_quote(buy_address, sell_address, amount)
        return await self.account.get_sign(swap_order)



"""
 DEX SWAP
🔒 USER RELATED
"""
from typing import Optional

from loguru import logger
from web3 import Web3

from dxsp import __version__
from dxsp.config import settings
from dxsp.utils.contract_utils import ContractUtils
from dxsp.utils.explorer_utils import get_account_transactions


class AccountUtils:

    """
    Class AccountUtils to interact with private related methods
    such as account balance, signing transactions, etc.

    Args:
        w3 (Optional[Web3]): Web3

    Methods:
        get_info()
        get_name()
        get_help()
        get_account_balance()
        get_trading_asset_balance()
        get_account_position
        get_account_margin
        get_account_open_positions
        get_account_transactions()
        get_account_pnl
        get_approve
        get_sign
        get_gas
        get_gas_price

    """

    def __init__(self, w3: Optional[Web3] = None):
        self.logger = logger
        self.w3 = w3 or Web3(Web3.HTTPProvider(settings.dex_rpc))
        self.wallet_address = self.w3.to_checksum_address(settings.dex_wallet_address)
        self.account_number = (
            f"{str(self.w3.net.version)} - " f"{str(self.wallet_address)[-8:]}"
        )
        self.private_key = settings.dex_private_key
        self.trading_asset_address = self.w3.to_checksum_address(
            settings.trading_asset_address
        )
        self.contract_utils = ContractUtils(w3=self.w3)
        self.commands = settings.dxsp_commands

    async def get_info(self):
        """
        Get the information about the DexSwap API.

        Returns:
            str: A string containing the version of DexSwap, the name obtained from
                 `get_name()`, and the account number.
        Raises:
            Exception: If there is an error while retrieving the information.
        """
        try:
            return (
                f"ℹ️ DexSwap v{__version__}\n"
                f"💱 {await self.get_name()}\n"
                f"🪪 {self.account_number}"
            )
        except Exception as error:
            return error

    async def get_name(self):
        """
        Retrieves the name of the object being the
        last 8 characters of the router contract address.

        :return: A string representing
        the name of the object.
        """
        if settings.dex_router_contract_addr:
            return str(settings.dex_router_contract_addr)[-8:]

    async def get_help(self):
        """
        Asynchronously retrieves the help information.

        Returns:
            str: The help information,
            including the available commands.
        """
        return f"{self.commands}\n"

    async def get_account_balance(self):
        """
        Retrieves the account balance of the user.

        Returns:
            str: A formatted string containing
            the account balance in Bitcoin (₿) and
            the trading asset balance like USDT (💵).
        """
        account_balance = self.w3.eth.get_balance(
            self.w3.to_checksum_address(self.wallet_address)
        )
        account_balance = self.w3.from_wei(account_balance, "ether") or 0
        trading_asset_balance = await self.get_trading_asset_balance()
        return f"₿ {round(account_balance,5)}\n💵 {trading_asset_balance}"

    async def get_trading_asset_balance(self):
        """
        Retrieves the balance of the trading asset
        for the current wallet address.

        Returns:
            The balance of the trading asset as a float.
            If the balance is not available,
            it returns 0.
        """
        trading_asset_balance = await self.contract_utils.get_token_balance(
            self.trading_asset_address, self.wallet_address
        )
        return trading_asset_balance if trading_asset_balance else 0

    async def get_account_position(self):
        """
        Retrieves the account position.

        Returns:
            str: A string representing the account position.
        """
        position = "📊 Position\n"
        position += f"Opened: {str(await self.get_account_open_positions())}\n"
        position += f"Margin: {str(await self.get_account_margin())}"
        return position

    async def get_account_margin(self):
        """
        Get the account margin. Not yet implemented

        Returns:
            int: The account margin.
        """
        return 0

    async def get_account_open_positions(self):
        """
        Get the open positions for the account.
        Not yet implemented

        :return: The number of open positions
        for the account.
        """
        return 0

    async def get_account_transactions(self, period=24):
        """
        Retrieve the account transactions for a given period.
        Not yet implemented

        Args:
            period (int): The time period in hours
            to retrieve the transactions for.
            Default is 24 hours.

        Returns:
            List[Transaction]: A list of transactions for the account.
        """
        return await get_account_transactions(
            period, self.trading_asset_address, self.wallet_address
        )

    async def get_account_pnl(self, period=24):
        """
        Create a profit and loss (PnL)
        report for the account.
        Not yet implemented

        Args:
            period (int): The time period in hours
            to retrieve the PnL for. Default is 24 hours.

        Returns:
            str: A string containing the PnL report.


        """
        pnl_dict = await self.get_account_transactions(period)
        pnl_report = "".join(
            f"{token} {value}\n" for token, value in pnl_dict["tokenList"].items()
        )
        pnl_report += f"Total {pnl_dict['pnl']}\n"
        pnl_report += await self.get_account_position()

        return pnl_report

    async def get_approve(self, token_address):
        """
        Given a token address, approve a token

        Args:
            token_address (str): The token address

        Returns:
            approval_tx_hash

        """
        try:
            contract = await self.contract_utils.get_token_contract(token_address)
            if contract is None:
                return
            approved_amount = self.w3.to_wei(2**64 - 1, "ether")
            owner_address = self.w3.to_checksum_address(self.wallet_address)
            dex_router_address = self.w3.to_checksum_address(
                settings.dex_router_contract_addr
            )
            allowance = contract.functions.allowance(
                owner_address, dex_router_address
            ).call()
            if allowance == 0:
                approval_tx = contract.functions.approve(
                    dex_router_address, approved_amount
                )
                approval_tx_hash = await self.get_sign(approval_tx.transact())
                return self.w3.eth.wait_for_transaction_receipt(approval_tx_hash)
        except Exception as error:
            raise ValueError(f"Approval failed {error}")

    async def get_sign(self, transaction):
        """
        Given a transaction, sign a transaction

        Args:
            transaction (Transaction): The transaction

        Returns:
            signed_tx_hash

        """
        try:
            signed_tx = self.w3.eth.account.sign_transaction(
                transaction, self.private_key
            )
            return self.w3.eth.send_raw_transaction(signed_tx.rawTransaction)
        except Exception as error:
            raise error

    async def get_gas(self, transaction):
        """
        Given a transaction, get gas estimate

        Args:
            transaction (Transaction): The transaction

        Returns:
            int: The gas estimate

        """
        gas_limit = self.w3.eth.estimate_gas(transaction) * 1.25
        return int(self.w3.to_wei(gas_limit, "wei"))

    async def get_gas_price(self):
        """
        search get gas price

        Returns:
            int: The gas price

        """
        return round(self.w3.from_wei(self.w3.eth.generate_gas_price(), "gwei"), 2)
